/**********************************************************/
/*    BATTLE:  PLAY BATTLESHIP                            */
/*                                                        */
/*     PETER D. MATTHEWS, JR.               1976          */
/*               REVISED         1977, SEPT 1981          */
/*               REVISED FOR IBM PC     DEC 1983          */
/*               Converted to C         OCT 1984          */
/**********************************************************/

#include "stdio.h"

#define ESC   27
#define CR    13
#define BEL    7
#define BS     8
#define SPACE 32
#define LEFT  (-75)
#define RIGHT (-77)

/**********************************************************/
/*   battle:  declarations  (all dimensions +1 for C)     */
/**********************************************************/

static int   test = FALSE;
static char  exact, near;  /* results returned by "eval" */
static int   nturn;

static int  nsink[5];
       int  ipos[9][9];   
       int  iset[10][10]; /* overdim for crude touch alg. */
static char lasts[9][28];
static int  fight;
static int  nshots;
static int  nhit[5];
static int  nehit[5];
static int  hturn[5][6];
static int  neshots;

static float epoints;
static float points;
static float eff;

static int  ninfo = 12;
static char info[12][72] = { " ",
      "\nSound General Quarters!!!\n",
      "Elements of the invading enemy navy are approaching.",
      "The enemy squadron is of the SAME STRENGTH as yours.",
      "As admiral of our squadron, you have:\n",
      "     1 Battleship    (BB)    5 sectors   2 shots",
      "     1 Heavy Cruiser (CA)    4 sectors   1 shot",
      "     1 Light Cruiser (CL)    3 sectors   1 shot",
      "     1 Destroyer     (DD)    2 sectors   1 shot\n",
      "The enemy is known to be of reasonable intelligence and of",
      "great determination.  This will be a fight to the death!\n",
      "Clear for Action!!!\n"
      };
static char name[5][14] = {" ",
      "Battleship",  "Heavy Cruiser",  "Light Cruiser",  "Destroyer"
      };
static char sep[] =   "  +----+----+----+----+----+----+----+----+";
static char tline[] = "    A    B    C    D    E    F    G    H";
static int  ishot[6][3], ish1, ish2;

static int  nshots1;
static char letter8[] = " ABCDEFGH";
static char number8[] = " 12345678";

static int  nflash = 6;
static char flash[6][72] = { " ",
      "BATTLESHIP                        Pete Matthews   (c) 1984\n\n",
      "FLASH!  The gyro-aided gunsight is now available!\n",
      "If you choose to use it, it is easier to win.\n",
      "If you choose not to use the gyro-aided gunsight, then you",
      "should still be able to win regularly, with good play.\n"
      };

static int iadjust,touch,ir,notir;
static int notshot, notshot1, irhit[5], irand, iex;


/**********************************************************/
/*   battle:  begin executable code                       */
/**********************************************************/

main (argc, argv)
int  argc;
char *argv[];

{
  int i;

  if (argc>1)
    test = TRUE;
  
  scr_setup();  /* set up for routines in PCIO.A */

  PLAY:
  clear ();
  for (i=1;i<nflash;i++)
    printf ("%s\n", flash[i]);
  initial();

  /* 2==> smart enemy wont shoot every box (leaves one hole per turn) */
  if (yes_no(3))  iadjust = 1;
  else  iadjust = 2;
  co('\n');
  /* allow touching ships? */
  if (touch = yes_no(5))  iadjust+=3;
  
  clear ();
  for (i=1;i<ninfo;i++)
    printf ("%s\n", info[i]);
  eset();

  for (nturn=1;fight;nturn++)
    {
    getshot();
    enemy();
    myshot();
    }
  grid();

  if (yes_no(1)) /* game over - play again?*/ 
    goto  PLAY;
  exit(0);  
}


/**********************************************************/
/**   eset:  set up ships to shoot at                     */
/**********************************************************/

eset()
{
int i,j,k,l,n, irnd[3], irnd2[3];

for (i=1;i<=4;i++)
  {
  n = 9 - nsink[i];

  retry:

  /* determine random placement */
  ir = roll(2);
  if (ir==1)  notir = 2; 
  else  notir=1;
  irnd[ir] = roll(n);
  irnd[notir] = roll(8);
  irnd2[1] = irnd[1];
  irnd2[2] = irnd[2];

  /* see if ship fits */
  for (j=1;j<=nsink[i];j++)
    {
    k = irnd2[1];
    l = irnd2[2];
    if (iset[k][l]!=0)  goto retry;
    irnd2[ir] += 1;
    }

  /* mark array to place ship */
  for (j=1;j<=nsink[i];j++)
    {
    k = irnd[1];
    l = irnd[2];
    iset[k][l] = i;
    irnd[ir] += 1;
    if (!touch)  /* mark adjacent (crude - all 8) */
      {
      if (iset[--k][--l]==0)  iset[k][l] = 9;
      if (iset[k][++l]==0)    iset[k][l] = 9;
      if (iset[k][++l]==0)    iset[k][l] = 9;
      if (iset[++k][l]==0)    iset[k][l] = 9;
      if (iset[++k][l]==0)    iset[k][l] = 9;
      if (iset[k][--l]==0)    iset[k][l] = 9;
      if (iset[k][--l]==0)    iset[k][l] = 9;
      if (iset[--k][l]==0)    iset[k][l] = 9;
      }
    }
  }

  if (!touch)  /* turn off touch indicators */ 
    for (i=1;i<=8;i++)
      for (j=1;j<=8;j++)
        if (iset[i][j]==9)  iset[i][j] = 0;;

  return;
}


/**********************************************************/
/*   grid:  print grid                                    */
/**********************************************************/

grid ()
  {
  int i, j, k, l;

  if (!fight)  /* show ships as "-" */ 
    {
    printf("\nShip locations will be \"-\"");
    for (i=1;i<=8;i++)
      for (j=1;j<=8;j++)
        {
        l = iset[i][j];
        if (l>0)  ipos[i][j] = -99;
        else  if (l<0)  ipos[i][j] = -ipos[i][j];
        }
    }

  printf ("\nPress any key to continue");
  l = inkey();
  clear ();
  printf (" %s", tline);
  for (i=1;i<=8;i++)
    {
    printf ("\n %s", sep);
    printf ("\n %-2d:", i);
    for (j=1;j<=8;j++)
      {
      l = ipos[i][j];
      if (l==0)  printf("    :");
      else  printf ("%3d :", l);
      }
    printf ("%s", lasts[i]);
    k = i - 1;
    if (k>=1 && k<=4)
      for (j=1;j<=5;j++)
        {
        l = hturn[k][j];
        if (l>0)  printf ("%3d",l);
        }
    else  if (i==6) 
      for(j=1;j<=4;j++)
        printf ("%3d", nehit[j]);
    else  if (i==7)
        printf ("%9d", nturn);
    else  if (i==8)
        printf ("%9.1f", eff);
    }
  printf ("\n %s\n", sep);
  return;
  }


/**********************************************************/
/*   getshot:  get shots from keyboard                    */
/**********************************************************/

getshot()
  {
  int i,j,k;

  gs0:
  i = 1;
  gs1:
    grid();
    if (i>1)  aimed(i);
    k = nshots+1-i;
    printf("\nAim %d shots:  ", k);

    gs2:
      getone();
      if (ish1>0)
        {
        ishot[i][1] = ish1;
        ishot[i][2] = ish2;
        if (++i<=nshots)  goto gs2;
        }
      else if(ish1<0 && i>1)
        {
        i--;
        aimed(i);
        goto gs2;
        }

    if (i<=nshots)  goto gs1;
 
  aimed(i);
  printf ("\nFire (Y or N)?");
  gs3:
  k = inkey();
  if (k=='n' || k=='N') 
    goto gs0;
  if (k!='y' && k!='Y' && k!=CR) 
    {
    co(BEL);
    goto gs3;
    }
  return;
  }


/**********************************************************/
/*   getone:  get one shot (results in ish1, ish2)        */
/*            ish1 = 0 if CR pressed when no partial shot */
/*            ish1 =-1 if BS pressed when no partial shot */
/*            exit possible if ESC pressed                */
/**********************************************************/

getone() 
  {
  int i,j,k;

  go0:
  ish1 = 0;
  ish2 = 0;
  j = 0;
  go1:
    if (j==2)  
      {
      if (ipos[ish1][ish2]==0)  return;
      dmsg(6);  /* already shot */
      goto go0;
      } 
    k = inkey();
    if (k==BS)
      {  
      if (j==0)  
        {
        ish1 = -1;
        return;  
        }
      co('|');
      goto go0;
      }  
    if (k==CR && j==0)  return;  
    if (k==' ')  
      {
      co(' ');
      goto go1;  
      }  
    if (k==ESC)
      {     
      if (yes_no(2))
        {
        fight = FALSE;
        grid();
        exit(0); /* END GAME */
        }
      co (BEL);
      goto go1;
    }
  
    k = toupper(k);
    for (i=1;i<=8;i++)
      {
      if (k==number8[i] && ish1==0)
        {
        ish1 = i;
        co(k);
        j++;
        goto go1;
        }
      if (k==letter8[i] && ish2==0)
        {
        ish2 = i;
        co(k);
        j++;
        goto go1;
        }
      }     
    co(BEL);
    goto go1;
  }


/**********************************************************/
/*   aimed:  print shots which have been aimed            */
/**********************************************************/
aimed(n) 
int n;
  {
  int i,j,k;
  printf ("\n%-1d/%-1d shots aimed:", --n, nshots); 
  for (i=1;i<=n;i++)
    {
    j = ishot[i][2];
    k = ishot[i][1];
    printf ("  %c%c", letter8[j], number8[k]);
    }
  return;
  }

/**********************************************************/
/*   enemy:  simulate enemy turn                          */
/**********************************************************/

enemy ()
  {
  int i, j, k;

  nshots1 = nshots;
  if (iex>0)  iex--;
  ircomp();

  for (j=1;j<=neshots;j++)
    { 
    notshot1 = notshot;
    if (notshot1<irhit[4])  notshot1 = irhit[4] + iex;
    irand = roll(notshot1);
   
    if (j==1)  notshot -= iadjust;
    else  notshot--;
    if (irand>irhit[4])  printf ("\n%d Enemy shot missed", j);
    else  for(i=1;i<=4;i++) 
      {   /* enemy hit */
      if (irand<=irhit[i]) 
        {
        nehit[i] += 1;
        printf ("\n%d Enemy shot hit your %s", j, name[i]);
        if (nehit[i]==nsink[i]) 
          {
          beep(1);
          printf ("\n    * %s sunk *", name[i]);
          if (i==1)
            {
            beep(1);
            nshots--; /* battleship */
            }
          nshots--;
          ircomp();
          }
        goto endhit;
        }
      }  /* enemy hit */
    endhit:
    if (nshots==0)
      {
      beep(4);
      printf ("\n   ***  All of your ships are sunk  ***");
      fight = FALSE;
      goto endenemy;
      }
    }

  endenemy: 
  k = 2*nehit[1]+nehit[2]+nehit[3]+nehit[4]+10 - 2*nshots;
  epoints = k;
  return;
  }


/**********************************************************/
/*    ircomp:  compute random hit limits                  */
/**********************************************************/
ircomp()
  { 
  int i;

  if (test) 
    printf ("\n\nNote:  on a %d-sided die, I hit BB,CA,CL,DD with ", notshot);

  for (i=1;i<=4;i++)
    {
    if (nehit[i]<nsink[i])
      {
      irhit[i] = irhit[i-1] + nsink[i];
      if (nehit[i]>0) irhit[i] += nehit[i];
      }
    else  irhit[i] = irhit[i-1];
    if (test) 
      printf(" %d", irhit[i]);
    }
  return;
  }


/**********************************************************/
/*   myshot:  calculate results of player turn            */
/**********************************************************/
myshot ()
  {
  int  nhtt[5] = {0,0,0,0,0} ;/* hits in turn */
  int  miss = 0;
  int  i,i1,i2,ih,l1,l2,k;

  if (test)
    printf("\n Results in order: ");
  for (i=1;i<=nshots1;i++)
    {
    i1 = ishot[i][1];
    i2 = ishot[i][2];
    ipos[i1][i2] = nturn;
    ih = iset[i1][i2];
    if (ih>0)
      {
      iset[i1][i2] = -ih;
      nhit[ih] += 1;
      nhtt[ih] += 1;
      k = nhit[ih];
      hturn[ih][k] = nturn;
      }
    else  miss++;
    if (test)
      printf(" %d", ih);  
    }

  co ('\n');  
  for (i=1;i<=4;i++)
    if (nhtt[i]>0)
      {
      printf ("\n%d hits on enemy %s", nhtt[i], name[i]);
      if (nhit[i]==nsink[i]) 
        {
        beep(1); 
        printf ("\n    * %s sunk *", name[i]);
        if (i==1)
          {
          beep(1);
          neshots--; /* battleship */
          }
        neshots--;
        }
      }
    
  printf ("\n%d shots missed\n", miss);

  if (neshots==0)
    {
    beep(4);
    printf ("\n\n   ***  All enemy ships are sunk  ***");
    fight = FALSE;
    }

  i = 2*nhit[1] + nhit[2] + nhit[3] + nhit[4] + 10 - 2*neshots;
  points = i;
  eff = points + epoints;
  if (eff>0.0)  eff = 1000.0 * (points-epoints) / eff;
  else eff = 0.0;
  return;
  }
 

/**********************************************************/
/*   beep:   routine to beep n times                      */
/**********************************************************/
beep (n)
int n;
  {
  int i,j;
  for (i=1;i<=n;i++)
    {
    j=1;
    while(j<5000)  j++;
    co(BEL);
    }
  return;
  }


/**********************************************************/
/*   ROLL:  RETURNS RANDOM INTEGER FROM 1 TO N            */
/**********************************************************/
roll (n) 
int  n;

{
         int        iy1;
         double     fr, frand();
  static int        first = TRUE;
         int        timer[2];


  /*  INITIALIZE ON FIRST PASS  */
  if (first==TRUE) 
  {
    timer[0] = _peek (0x6c, 0x40); /* low word of PC timer (offset, segment) */
    timer[1] = _peek (0x6d, 0x40); 
    iy1 = (128*timer[0]) + timer[1];
    srand (iy1); /* seed random nbr generator */  
    first = FALSE;
  }
  
  fr = frand();
  iy1 = (int) (fr * n);
  iy1 %= n;
  return (++iy1);
}



/**********************************************************/
/*   yes_no:  question?                                   */
/**********************************************************/
yes_no(n)
int n;
{
  int c;
  
  ask:
  dmsg(n);
  c = inkey();
  if ((c==ESC) && (n!=2))
    exit(0); /* not if quit msg */
  if (c=='Y' || c=='y')
    return (TRUE);
  if (c=='N' || c=='n')
    return (FALSE);
  co(BEL);
  goto ask;
}


/**********************************************************/
/**   dmsg:  display message by number                    */
/**********************************************************/

dmsg (n)

int n;

{
  if (n==1)
    printf ("\nGAME OVER - Play Again (Y or N)?      ");
  else if (n==2)	  
    printf ("\nQuit (Y or N)?                        ");
  else if (n==3)	
    printf ("\nWill you use the gyro-aided gunsight (Easy game - Y or N)? ");
  else if (n==4)  /* not used - rely on ANSI.SYS */	
    printf ("\nDo you have a color display (Y or N)? ");
  else if (n==5) 	
    printf ("\nAllow touching ships (Y or N)?        ");
  else if (n==6) 	
    printf ("\nYou already shot there!   Aim again:  ");
  return;
} 


/**********************************************************/
/*   inkey:  get input keystroke, extended codes negative */
/**********************************************************/

inkey ()
{
  int k;
  k = ci ();
  if (k==0)
    k = - ci ();
  return(k);
}


/*********************************************************/
/*  initial:  initialize values for replay               */
/*********************************************************/


initial ()
{
  int i,j;
  static char author[] = " Pete Matthews  1984 ";

  for (i=0;i<=8;i++)
    for (j=0;j<=8;j++)
      {
      ipos[i][j] = 0;
      iset[i][j] = 0;
      }

  for (i=0;i<=4;i++)
    {
    nhit[i] = 0;
    nehit[i] = 0;
    irhit[i] = 0;
    for (j=0;j<=5;j++)
      hturn[i][j] = 0;
    }

  nsink[1] = 5;
  nsink[2] = 4;
  nsink[3] = 3;
  nsink[4] = 2;
  strncpy (lasts[1], "   RECORDED HITS BY TURN:", 28);
  strncpy (lasts[2], "   BB (5):  ", 28);
  strncpy (lasts[3], "   CA (4):  ", 28);
  strncpy (lasts[4], "   CL (3):  ", 28);
  strncpy (lasts[5], "   DD (2):  ", 28);
  strncpy (lasts[6], "   Enemy Hits: ", 28);
  strncpy (lasts[7], "   Turn Number: ", 28);
  strncpy (lasts[8], "   Command Rating:", 28);
  fight = TRUE;
  nshots = 5;
  neshots = 5;
  epoints = 0;
  points = 0;
  eff = 0;
  notshot = 64;
  iex = 15; 
  return;
}


/*********************************************************/
/*   clear:  clear screen and home cursor                */
/*********************************************************/

clear ()
  {
  scr_clr();
  scr_rowcol(0,0);
  return;
  }
